/******************************************************************************
 * Copyright (c) 2011-2023, NVIDIA CORPORATION.  All rights reserved.
 * Modifications Copyright (c) 2024-2025, Advanced Micro Devices, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the NVIDIA CORPORATION nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NVIDIA CORPORATION BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 ******************************************************************************/

#pragma once

// Benchmark utils
#include "../../bench_utils/bench_utils.hpp"

// rocThrust
#include <thrust/device_vector.h>
#include <thrust/execution_policy.h>
#include <thrust/set_operations.h>
#include <thrust/sort.h>

// Google Benchmark
#include <benchmark/benchmark.h>

// STL
#include <cstddef>
#include <string>
#include <vector>

struct by_key
{
  template <typename KeyT, typename ValueT, typename OpT, typename Policy>
  float64_t
  run(thrust::device_vector<KeyT>& in_keys,
      thrust::device_vector<ValueT>& in_vals,
      thrust::device_vector<KeyT>& out_keys,
      thrust::device_vector<ValueT>& out_vals,
      const std::size_t elements_in_A,
      const OpT op,
      Policy policy)
  {
    bench_utils::gpu_timer d_timer;

    d_timer.start(0);
    op(policy,
       in_keys.cbegin(),
       in_keys.cbegin() + elements_in_A,
       in_keys.cbegin() + elements_in_A,
       in_keys.cend(),
       in_vals.cbegin(),
       in_vals.cbegin() + elements_in_A,
       out_keys.begin(),
       out_vals.begin());
    d_timer.stop(0);

    return d_timer.get_duration();
  }
};

template <class KeyT, class ValueT, class OpT>
void run_benchmark(benchmark::State& state,
                   const std::size_t elements,
                   const std::string seed_type,
                   const int entropy_reduction,
                   const std::size_t input_size_ratio)
{
  // Benchmark object
  by_key benchmark{};

  // GPU times
  std::vector<double> gpu_times;

  // Generate input
  const auto entropy       = bench_utils::get_entropy_percentage(entropy_reduction) / 100.0f;
  const auto elements_in_A = static_cast<std::size_t>(static_cast<double>(input_size_ratio * elements) / 100.0f);

  thrust::device_vector<KeyT> in_keys = bench_utils::generate(elements, seed_type, entropy);
  thrust::device_vector<KeyT> out_keys(elements);

  thrust::device_vector<ValueT> in_vals(elements);
  thrust::device_vector<ValueT> out_vals(elements);

  thrust::sort(in_keys.begin(), in_keys.begin() + elements_in_A);
  thrust::sort(in_keys.begin() + elements_in_A, in_keys.end());

  OpT op{};
  bench_utils::caching_allocator_t alloc{};
  thrust::detail::device_t policy{};
  // not a warm-up run, we need to run once to determine the size of the output
  auto result_ends = op(
    policy(alloc),
    in_keys.cbegin(),
    in_keys.cbegin() + elements_in_A,
    in_keys.cbegin() + elements_in_A,
    in_keys.cend(),
    in_vals.cbegin(),
    in_vals.cbegin() + elements_in_A,
    out_keys.begin(),
    out_vals.begin());
  const std::size_t elements_in_AB = thrust::distance(out_keys.begin(), result_ends.first);

  for (auto _ : state)
  {
    float64_t duration =
      benchmark.template run<KeyT, ValueT>(in_keys, in_vals, out_keys, out_vals, elements_in_A, op, policy(alloc));
    state.SetIterationTime(duration);
    gpu_times.push_back(duration);
  }

  // BytesProcessed include read and written bytes, so when the BytesProcessed/s are reported
  // it will actually be the global memory bandwidth gotten.
  const std::size_t global_memory_key_bytes   = (elements + elements_in_AB) * sizeof(KeyT);
  const std::size_t global_memory_value_reads = OpT::read_all_values ? elements : elements_in_A;
  const std::size_t global_memory_value_bytes = (global_memory_value_reads + elements_in_AB) * sizeof(ValueT);
  state.SetBytesProcessed(state.iterations() * (global_memory_key_bytes + global_memory_value_bytes));
  state.SetItemsProcessed(state.iterations() * elements);

  const double gpu_cv         = bench_utils::StatisticsCV(gpu_times);
  state.counters["gpu_noise"] = gpu_cv;
}

#define CREATE_BENCHMARK(KeyT, ValueT, Elements, EntropyReduction, InputSizeRatio)                                    \
  benchmark::RegisterBenchmark(                                                                                       \
    bench_utils::bench_naming::format_name(                                                                           \
      "{algo:" + algo_name + ",subalgo:by_key" + ",key_type:" #KeyT + ",value_type:" #ValueT + ",elements:" #Elements \
      + ",entropy:" + std::to_string(bench_utils::get_entropy_percentage(EntropyReduction))                           \
      + ",input_size_ratio:" #InputSizeRatio)                                                                         \
      .c_str(),                                                                                                       \
    run_benchmark<KeyT, ValueT, OpT>,                                                                                 \
    Elements,                                                                                                         \
    seed_type,                                                                                                        \
    EntropyReduction,                                                                                                 \
    InputSizeRatio)

#define BENCHMARK_ELEMENTS(key_type, value_type, elements, entropy) \
  CREATE_BENCHMARK(key_type, value_type, elements, entropy, 25),    \
    CREATE_BENCHMARK(key_type, value_type, elements, entropy, 50),  \
    CREATE_BENCHMARK(key_type, value_type, elements, entropy, 75)

#define BENCHMARK_VALUE_TYPE(key_type, value_type, entropy)     \
  BENCHMARK_ELEMENTS(key_type, value_type, 1 << 16, entropy),   \
    BENCHMARK_ELEMENTS(key_type, value_type, 1 << 20, entropy), \
    BENCHMARK_ELEMENTS(key_type, value_type, 1 << 24, entropy), \
    BENCHMARK_ELEMENTS(key_type, value_type, 1 << 28, entropy)

#define BENCHMARK_KEY_TYPE_ENTROPY(key_type, entropy) \
  BENCHMARK_VALUE_TYPE(key_type, int8_t, entropy), BENCHMARK_VALUE_TYPE(key_type, int64_t, entropy)

template <class OpT>
void add_benchmarks(
  const std::string& algo_name, std::vector<benchmark::internal::Benchmark*>& benchmarks, const std::string seed_type)
{
  constexpr int entropy_reductions[] = {0, 4}; // 1.000, 0.201;

  for (int entropy_reduction : entropy_reductions)
  {
    std::vector<benchmark::internal::Benchmark*> bs = {
      BENCHMARK_KEY_TYPE_ENTROPY(int8_t, entropy_reduction),
      BENCHMARK_KEY_TYPE_ENTROPY(int16_t, entropy_reduction),
      BENCHMARK_KEY_TYPE_ENTROPY(int32_t, entropy_reduction),
      BENCHMARK_KEY_TYPE_ENTROPY(int64_t, entropy_reduction)};

    benchmarks.insert(benchmarks.end(), bs.begin(), bs.end());
  }
}
